@@ -42,7 +42,6 @@
 #include <boost/bimap.hpp>
 #include <boost/bimap/multiset_of.hpp>
 #include <boost/bimap/set_of.hpp>
-#include <boost/property_map/property_map.hpp>
 #include <boost/range.hpp>
 #include <boost/unordered_map.hpp>
 #include <boost/unordered_set.hpp>
@@ -57,14 +56,17 @@
 #ifdef CGAL_PMP_REMESHING_DEBUG
 #include <CGAL/Polygon_mesh_processing/self_intersections.h>
 #define CGAL_DUMP_REMESHING_STEPS
+#define CGAL_PMP_REMESHING_VERBOSE_PROGRESS
 #endif
 
 #ifdef CGAL_PMP_REMESHING_VERY_VERBOSE
 #define CGAL_PMP_REMESHING_VERBOSE
+#define CGAL_PMP_REMESHING_VERBOSE_PROGRESS
 #endif
 
-
-
+#ifdef CGAL_PMP_REMESHING_VERBOSE_PROGRESS
+#define CGAL_PMP_REMESHING_VERBOSE
+#endif
 
 
 namespace CGAL {
@@ -146,6 +148,49 @@ namespace internal {
     }
   };
 
+  template <typename PM,
+            typename EdgeIsConstrainedMap>
+  struct Connected_components_pmap
+  {
+    typedef typename boost::graph_traits<PM>::face_descriptor   face_descriptor;
+    typedef std::size_t                                         Patch_id;
+    typedef Connected_components_pmap<PM, EdgeIsConstrainedMap> CCMap;
+
+    boost::unordered_map<face_descriptor, Patch_id> patch_ids_map;
+
+  public:
+    typedef face_descriptor                     key_type;
+    typedef Patch_id                            value_type;
+    typedef Patch_id&                           reference;
+    typedef boost::read_write_property_map_tag  category;
+
+    Connected_components_pmap()
+      : patch_ids_map()
+    {}
+    Connected_components_pmap(const PM& pmesh, EdgeIsConstrainedMap ecmap)
+      : patch_ids_map()
+    {
+      PMP::connected_components(pmesh,
+        boost::make_assoc_property_map(patch_ids_map),
+        PMP::parameters::edge_is_constrained_map(ecmap));
+    }
+
+    friend value_type get(const CCMap& m, const key_type& f)
+    {
+      CGAL_assertion(!m.patch_ids_map.empty());
+      CGAL_assertion(m.patch_ids_map.find(f) != m.patch_ids_map.end());
+
+      return m.patch_ids_map.at(f);
+    }
+    friend void put(CCMap& m, const key_type& f, const value_type i)
+    {
+      if (m.patch_ids_map.find(f) == m.patch_ids_map.end())
+        m.patch_ids_map.insert(std::make_pair(f, i));
+      else
+        m.patch_ids_map[f] = i;
+    }
+  };
+
   template<typename PM,
            typename EdgeConstraintMap,
            typename VertexPointMap>
@@ -179,6 +224,8 @@ namespace internal {
               typename boost::graph_traits<PolygonMesh>::edge_descriptor>
          , typename VertexIsConstrainedMap = No_constraint_pmap<
               typename boost::graph_traits<PolygonMesh>::vertex_descriptor>
+         , typename FacePatchMap = Connected_components_pmap<
+              PolygonMesh, EdgeIsConstrainedMap>
   >
   class Incremental_remesher
   {
@@ -197,10 +244,11 @@ namespace internal {
                                , GeomTraits
                                , EdgeIsConstrainedMap
                                , VertexIsConstrainedMap
+                               , FacePatchMap
                                > Self;
 
   private:
-    typedef std::size_t                                  Patch_id;
+    typedef typename boost::property_traits<FacePatchMap>::value_type Patch_id;
     typedef std::vector<Triangle_3>                      Triangle_list;
     typedef std::vector<Patch_id>                        Patch_id_list;
 
@@ -213,8 +261,9 @@ namespace internal {
     Incremental_remesher(PolygonMesh& pmesh
                        , VertexPointMap& vpmap
                        , const bool protect_constraints
-                       , EdgeIsConstrainedMap ecmap = EdgeIsConstrainedMap()
-                       , VertexIsConstrainedMap vcmap = VertexIsConstrainedMap()
+                       , EdgeIsConstrainedMap ecmap
+                       , VertexIsConstrainedMap vcmap
+                       , FacePatchMap fpmap
                        , const bool own_tree = true)//built by the remesher
       : mesh_(pmesh)
       , vpmap_(vpmap)
@@ -223,7 +272,7 @@ namespace internal {
       , input_patch_ids_()
       , halfedge_status_map_()
       , protect_constraints_(protect_constraints)
-      , patch_ids_map_()
+      , patch_ids_map_(fpmap)
       , ecmap_(ecmap)
       , vcmap_(vcmap)
     {
@@ -240,12 +289,6 @@ namespace internal {
     void init_remeshing(const FaceRange& face_range)
     {
       tag_halfedges_status(face_range); //called first
-      Constraint_property_map cpmap(*this);
-
-      //build AABB tree of input surface
-      PMP::connected_components(mesh_,
-        boost::make_assoc_property_map(patch_ids_map_),// set patch_id() for each face
-        PMP::parameters::edge_is_constrained_map(cpmap));
 
       BOOST_FOREACH(face_descriptor f, face_range)
       {
@@ -384,7 +427,7 @@ namespace internal {
         double sqlen = eit->first;
         long_edges.right.erase(eit);
 
-#ifdef CGAL_PMP_REMESHING_VERBOSE
+#ifdef CGAL_PMP_REMESHING_VERBOSE_PROGRESS
         std::cout << "\r\t(" << long_edges.left.size() << " long edges, ";
         std::cout << nb_splits << " splits)";
         std::cout.flush();
@@ -393,6 +436,10 @@ namespace internal {
         if (protect_constraints_ && !is_longest_on_faces(edge(he, mesh_)))
           continue;
 
+        //collect patch_ids
+        Patch_id patch_id = get_patch_id(face(he, mesh_));
+        Patch_id patch_id_opp = get_patch_id(face(opposite(he, mesh_), mesh_));
+
         //split edge
         Point refinement_point = this->midpoint(he);
         halfedge_descriptor hnew = CGAL::Euler::split_edge(he, mesh_);
@@ -423,7 +470,6 @@ namespace internal {
         //insert new edges to keep triangular faces, and update long_edges
         if (!is_on_border(hnew))
         {
-          Patch_id patch_id = get_patch_id(face(hnew, mesh_));
           halfedge_descriptor hnew2 = CGAL::Euler::split_face(hnew,
                                                               next(next(hnew, mesh_), mesh_),
                                                               mesh_);
@@ -446,7 +492,6 @@ namespace internal {
         //do it again on the other side if we're not on boundary
         if (!is_on_border(hnew_opp))
         {
-          Patch_id patch_id = get_patch_id(face(hnew_opp, mesh_));
           halfedge_descriptor hnew2 = CGAL::Euler::split_face(prev(hnew_opp, mesh_),
                                                               next(hnew_opp, mesh_),
                                                               mesh_);
@@ -455,8 +500,8 @@ namespace internal {
             : MESH;
           halfedge_added(hnew2,                  snew);
           halfedge_added(opposite(hnew2, mesh_), snew);
-          set_patch_id(face(hnew2, mesh_), patch_id);
-          set_patch_id(face(opposite(hnew2, mesh_), mesh_), patch_id);
+          set_patch_id(face(hnew2, mesh_), patch_id_opp);
+          set_patch_id(face(opposite(hnew2, mesh_), mesh_), patch_id_opp);
 
           if (snew == PATCH)
           {
@@ -472,7 +517,6 @@ namespace internal {
 
 #ifdef CGAL_PMP_REMESHING_DEBUG
       CGAL_expensive_assertion(is_triangle_mesh(mesh_));
-      CGAL_assertion(halfedge_status_map_.size() == nb_valid_halfedges());
       debug_status_map();
       debug_self_intersections();
 #endif
@@ -518,7 +562,7 @@ namespace internal {
         halfedge_descriptor he = eit->second;
         short_edges.right.erase(eit);
 
-#ifdef CGAL_PMP_REMESHING_VERBOSE
+#ifdef CGAL_PMP_REMESHING_VERBOSE_PROGRESS
         std::cout << "\r\t(" << short_edges.left.size() << " short edges, ";
         std::cout << nb_collapses << " collapses)";
         std::cout.flush();
@@ -547,29 +591,32 @@ namespace internal {
         vertex_descriptor va = source(he, mesh_);
         vertex_descriptor vb = target(he, mesh_);
 
+        bool swap_done = false;
         if (is_on_patch_border(va) && !is_on_patch_border(vb))
         {
           he = opposite(he, mesh_);
           va = source(he, mesh_);
           vb = target(he, mesh_);
+          swap_done = true;
           CGAL_assertion(is_on_patch_border(vb) && !is_on_patch_border(va));
         }
-        else if (is_on_patch(va) && is_on_patch(vb))
-        {
-          if(!collapse_does_not_invert_face(he))
-          {
-            if (collapse_does_not_invert_face(opposite(he, mesh_)))
-            {
-              he = opposite(he, mesh_);
-              va = source(he, mesh_);
-              vb = target(he, mesh_);
-            }
-            else
-              continue;//both directions invert a face
-          }
-          CGAL_assertion(collapse_does_not_invert_face(he));
-        }
 
+        if(!collapse_does_not_invert_face(he))
+        {
+          if (!swap_done//if swap has already been done, don't re-swap
+           && collapse_does_not_invert_face(opposite(he, mesh_)))
+          {
+            he = opposite(he, mesh_);
+            va = source(he, mesh_);
+            vb = target(he, mesh_);
+
+            if (is_on_patch_border(va) && !is_on_patch_border(vb))
+              continue;//we cannot swap again. It would lead to a face inversion
+          }
+          else
+            continue;//both directions invert a face
+        }
+        CGAL_assertion(collapse_does_not_invert_face(he));
         CGAL_assertion(is_collapse_allowed(e));
 
         if (degree(va, mesh_) < 3
@@ -607,7 +654,8 @@ namespace internal {
           }
 
           //before collapse
-          bool mesh_border_case     = is_on_border(opposite(he, mesh_));
+          bool mesh_border_case     = is_on_border(he);
+          bool mesh_border_case_opp = is_on_border(opposite(he, mesh_));
           halfedge_descriptor ep_p  = prev(opposite(he, mesh_), mesh_);
           halfedge_descriptor epo_p = opposite(ep_p, mesh_);
           halfedge_descriptor en    = next(he, mesh_);
@@ -619,14 +667,25 @@ namespace internal {
 
           // merge halfedge_status to keep the more important on both sides
           //do it before collapse is performed to be sure everything is valid
-          merge_status(en, s_epo, s_ep);
           if (!mesh_border_case)
+            merge_status(en, s_epo, s_ep);
+          if (!mesh_border_case_opp)
             merge_status(en_p, s_epo_p, s_ep_p);
 
-          if (!mesh_border_case)
-            halfedge_and_opp_removed(prev(opposite(he, mesh_), mesh_));
           halfedge_and_opp_removed(he);
-          halfedge_and_opp_removed(prev(he, mesh_));
+          if (!mesh_border_case)
+            halfedge_and_opp_removed(prev(he, mesh_));
+          if (!mesh_border_case_opp)
+            halfedge_and_opp_removed(prev(opposite(he, mesh_), mesh_));
+
+          //constrained case
+          bool constrained_case = is_constrained(va) || is_constrained(vb);
+          if (constrained_case)
+          {
+            CGAL_assertion(is_constrained(va) ^ is_constrained(vb));//XOR
+            set_constrained(va, false);
+            set_constrained(vb, false);
+          }
 
           //perform collapse
           Point target_point = get(vpmap_, vb);
@@ -634,11 +693,13 @@ namespace internal {
           put(vpmap_, vkept, target_point);
           ++nb_collapses;
 
+          //fix constrained case
+          if (constrained_case)//we have made sure that collapse goes to constrained vertex
+            set_constrained(vkept, true);
+
           fix_degenerate_faces(vkept, short_edges, sq_low);
 
 #ifdef CGAL_PMP_REMESHING_DEBUG
-          CGAL_assertion_code(std::size_t nbb = nb_valid_halfedges());
-          CGAL_assertion(nbb == halfedge_status_map_.size());
           debug_status_map();
           CGAL_assertion(!incident_to_degenerate(halfedge(vkept, mesh_)));
 #endif
@@ -664,7 +725,6 @@ namespace internal {
 #endif
 
 #ifdef CGAL_PMP_REMESHING_DEBUG
-      CGAL_assertion(nb_valid_halfedges() == halfedge_status_map_.size());
       CGAL_expensive_assertion(is_triangle_mesh(mesh_));
       debug_status_map();
       debug_self_intersections();
@@ -687,10 +747,10 @@ namespace internal {
       BOOST_FOREACH(edge_descriptor e, edges(mesh_))
       {
         //only the patch edges are allowed to be flipped
-        halfedge_descriptor he = halfedge(e, mesh_);
-        if (!is_on_patch(he))
+        if (!is_flip_allowed(e))
           continue;
 
+        halfedge_descriptor he = halfedge(e, mesh_);
         vertex_descriptor va = source(he, mesh_);
         vertex_descriptor vb = target(he, mesh_);
         vertex_descriptor vc = target(next(he, mesh_), mesh_);
@@ -704,10 +764,12 @@ namespace internal {
         CGAL_assertion_code(Halfedge_status s1 = status(he));
         CGAL_assertion_code(Halfedge_status s1o = status(opposite(he, mesh_)));
 
+        Patch_id pid = get_patch_id(face(he, mesh_));
+
         CGAL::Euler::flip_edge(he, mesh_);
         ++nb_flips;
 
-#ifdef CGAL_PMP_REMESHING_VERBOSE
+#ifdef CGAL_PMP_REMESHING_VERBOSE_PROGRESS
         std::cout << "\r\t(" << nb_flips << " flips)";
         std::cout.flush();
 #endif
@@ -715,7 +777,6 @@ namespace internal {
         CGAL_assertion_code(Halfedge_status s2o = status(opposite(he, mesh_)));
         CGAL_assertion(s1 == s2   && s1 == PATCH);
         CGAL_assertion(s1o == s2o && s1o == PATCH);
-        CGAL_assertion(nb_valid_halfedges() == halfedge_status_map_.size());
         CGAL_assertion(!is_border(he, mesh_));
 
         CGAL_assertion(
@@ -748,6 +809,9 @@ namespace internal {
                (va == source(he, mesh_) && vb == target(he, mesh_))
             || (vb == source(he, mesh_) && va == target(he, mesh_)));
         }
+
+        set_patch_id(face(he, mesh_), pid);
+        set_patch_id(face(opposite(he, mesh_), mesh_), pid);
       }
 
 #ifdef CGAL_PMP_REMESHING_VERBOSE
@@ -755,7 +819,6 @@ namespace internal {
 #endif
 
 #ifdef CGAL_PMP_REMESHING_DEBUG
-      CGAL_assertion(nb_valid_halfedges() == halfedge_status_map_.size());
       debug_status_map();
       CGAL_assertion(0 == PMP::remove_degenerate_faces(mesh_
                             , PMP::parameters::vertex_point_map(vpmap_)
@@ -772,14 +835,20 @@ namespace internal {
     // "applies an iterative smoothing filter to the mesh.
     // The vertex movement has to be constrained to the vertex tangent plane [...]
     // smoothing algorithm with uniform Laplacian weights"
-    void tangential_relaxation(const bool smooth_along_features/*1d smoothing*/)
+    void tangential_relaxation(const bool relax_constraints/*1d smoothing*/
+                             , const unsigned int nb_iterations)
     {
-      //todo : move border vertices along 1-dimensional features
 #ifdef CGAL_PMP_REMESHING_VERBOSE
-      std::cout << "Tangential relaxation...";
-      std::cout.flush();
+      std::cout << "Tangential relaxation (" << nb_iterations << " iter.)...";
+      std::cout << std::endl;
+#endif
+      for (unsigned int nit = 0; nit < nb_iterations; ++nit)
+      {
+#ifdef CGAL_PMP_REMESHING_VERBOSE_PROGRESS
+        std::cout << "\r\t(iteration " << (nit + 1) << " / ";
+        std::cout << nb_iterations << ") ";
+        std::cout.flush();
 #endif
-
       //todo : use boost::vector_property_map to improve computing time
       typedef std::map<vertex_descriptor, Vector_3> VNormalsMap;
       VNormalsMap vnormals;
@@ -809,7 +878,7 @@ namespace internal {
 
           barycenters[v] = get(vpmap_, v) + move;
         }
-        else if (smooth_along_features
+        else if (relax_constraints
               && !protect_constraints_
               && is_on_patch_border(v)
               && !is_corner(v)
@@ -827,8 +896,8 @@ namespace internal {
           {
             vertex_descriptor ph0 = source(border_halfedges[0], mesh_);
             vertex_descriptor ph1 = source(border_halfedges[1], mesh_);
-            double dot = Vector_3(get(vpmap_, v), get(vpmap_, ph0))
-                       * Vector_3(get(vpmap_, v), get(vpmap_, ph1));
+            double dot = to_double(Vector_3(get(vpmap_, v), get(vpmap_, ph0))
+                                   * Vector_3(get(vpmap_, v), get(vpmap_, ph1)));
             //check squared cosine is < 0.25 (~120 degrees)
             if (0.25 < dot / (sqlength(border_halfedges[0]) * sqlength(border_halfedges[0])))
               barycenters[v] = CGAL::midpoint(midpoint(border_halfedges[0]),
@@ -853,12 +922,27 @@ namespace internal {
       // perform moves
       BOOST_FOREACH(const VP_pair& vp, new_locations)
       {
+        const Point initial_pos = get(vpmap_, vp.first);
+        const Vector_3 move(initial_pos, vp.second);
+
         put(vpmap_, vp.first, vp.second);
+
+        //check that no inversion happened
+        double frac = 1.;
+        while (frac > 0.03 //5 attempts maximum
+           && !check_normals(vp.first)) //if a face has been inverted
+        {
+          frac = 0.5 * frac;
+          put(vpmap_, vp.first, initial_pos + frac * move);//shorten the move by 2
+        }
+        if (frac <= 0.02)
+          put(vpmap_, vp.first, initial_pos);//cancel move
       }
 
       CGAL_assertion(is_valid(mesh_));
-
       CGAL_assertion(is_triangle_mesh(mesh_));
+      }//end for loop (nit == nb_iterations)
+
 #ifdef CGAL_PMP_REMESHING_DEBUG
       debug_self_intersections();
 #endif
@@ -870,23 +954,6 @@ namespace internal {
 #endif
     }
 
-    struct Constraint_property_map
-    {
-      typedef boost::readable_property_map_tag      category;
-      typedef bool                                  value_type;
-      typedef bool                                  reference;
-      typedef edge_descriptor                       key_type;
-      const Self* remesher_ptr_;
-      Constraint_property_map(): remesher_ptr_(NULL) {}
-      Constraint_property_map(const Self& remesher)
-        : remesher_ptr_(&remesher) {}
-      friend bool get(const Constraint_property_map& m,
-                      const edge_descriptor& e) {
-        return m.remesher_ptr_->is_on_border(e) ||
-               m.remesher_ptr_->is_on_patch_border(e);
-      }
-    };
-
     // PMP book :
     // "maps the vertices back to the surface"
     void project_to_surface()
@@ -899,7 +966,7 @@ namespace internal {
 
       BOOST_FOREACH(vertex_descriptor v, vertices(mesh_))
       {
-        if (!is_on_patch(v) || is_constrained(v))
+        if (is_constrained(v) || !is_on_patch(v))
           continue;
         //note if v is constrained, it has not moved
 
@@ -912,17 +979,13 @@ namespace internal {
           tree_ptr_->traits(),
           pid_pmap
           );
-
         tree_ptr_->traversal(get(vpmap_, v), projection_traits);
         CGAL_assertion(projection_traits.found());
-
         Point proj = projection_traits.closest_point();
         put(vpmap_, v, proj);
       }
-
       CGAL_assertion(is_valid(mesh_));
       CGAL_assertion(is_triangle_mesh(mesh_));
-
 #ifdef CGAL_PMP_REMESHING_DEBUG
       debug_self_intersections();
 #endif
@@ -938,18 +1001,14 @@ namespace internal {
 private:
   Patch_id get_patch_id(const face_descriptor& f) const
   {
-    if (patch_ids_map_.empty())
-      return -1;
-    else
-      return patch_ids_map_.at(f);
+    if (f == boost::graph_traits<PM>::null_face())
+      return Patch_id(-1);
+    return get(patch_ids_map_, f);
   }
 
   void set_patch_id(const face_descriptor& f, const Patch_id& i)
   {
-    if (patch_ids_map_.find(f) == patch_ids_map_.end())
-      patch_ids_map_.insert(std::make_pair(f, i));
-    else
-      patch_ids_map_[f] = i;
+    put(patch_ids_map_, f, i);
   }
 
   struct Patch_id_property_map
@@ -993,7 +1052,7 @@ private:
     double sqlength(const vertex_descriptor& v1,
                     const vertex_descriptor& v2) const
     {
-      return CGAL::squared_distance(get(vpmap_, v1), get(vpmap_, v2));
+      return to_double(CGAL::squared_distance(get(vpmap_, v1), get(vpmap_, v2)));
     }
 
     double sqlength(const halfedge_descriptor& h) const
@@ -1054,7 +1113,7 @@ private:
 
     bool is_constrained(const edge_descriptor& e) const
     {
-      return get(Constraint_property_map(*this), e);
+      return is_on_border(e) || is_on_patch_border(e);
     }
 
     bool is_split_allowed(const edge_descriptor& e) const
@@ -1086,10 +1145,12 @@ private:
         return false;
       if (is_on_patch(he)) //hopp is also on patch
       {
+        CGAL_assertion(is_on_patch(hopp));
         if (is_on_patch_border(target(he, mesh_)) && is_on_patch_border(source(he, mesh_)))
           return false;//collapse would induce pinching the selection
         else
-          return is_collapse_allowed(he) && is_collapse_allowed(hopp);
+          return (is_collapse_allowed_on_patch(he)
+               && is_collapse_allowed_on_patch(hopp));
       }
       else if (is_on_patch_border(he))
         return is_collapse_allowed_on_patch_border(he);
@@ -1098,7 +1159,7 @@ private:
       return false;
     }
 
-    bool is_collapse_allowed(const halfedge_descriptor& he) const
+    bool is_collapse_allowed_on_patch(const halfedge_descriptor& he) const
     {
       halfedge_descriptor hopp = opposite(he, mesh_);
 
@@ -1135,6 +1196,8 @@ private:
       {
         if (is_on_patch_border(next(hopp, mesh_)) && is_on_patch_border(prev(hopp, mesh_)))
           return false;
+        else if (next_on_patch_border(h) == hopp)
+          return false; //isolated patch border
         else
           return true;
       }
@@ -1142,6 +1205,24 @@ private:
       return true;//we already checked we're not pinching a hole in the patch
     }
 
+    bool is_flip_allowed(const edge_descriptor& e) const
+    {
+      return is_flip_allowed(halfedge(e, mesh_))
+          && is_flip_allowed(opposite(halfedge(e, mesh_), mesh_));
+    }
+
+    bool is_flip_allowed(const halfedge_descriptor& h) const
+    {
+      if (!is_on_patch(h))
+        return false;
+      if (!is_on_patch_border(target(h, mesh_)))
+        return true;
+      if ( is_on_patch_border(next(h, mesh_))
+        && is_on_patch_border(prev(opposite(h, mesh_), mesh_)))
+        return false;
+      return true;
+    }
+
     halfedge_descriptor next_on_patch_border(const halfedge_descriptor& h) const
     {
       CGAL_precondition(is_on_patch_border(h));
@@ -1198,37 +1279,43 @@ private:
       //move source at target
       put(vpmap_, vs, get(vpmap_, vt));
 
+      //collect halfedges around vs and vt
+      std::vector<halfedge_descriptor> hedges;
+      BOOST_FOREACH(halfedge_descriptor hd,
+        halfedges_around_target(h, mesh_))
+          hedges.push_back(hd);
+      BOOST_FOREACH(halfedge_descriptor hd,
+        halfedges_around_target(opposite(h, mesh_), mesh_))
+          hedges.push_back(hd);
+
       //collect normals to faces around vs AND vt
       //vertices are at the same location, but connectivity is still be same,
       //with plenty of degenerate triangles (which are common to both stars)
-      std::vector<Vector_3> normals;
-      BOOST_FOREACH(halfedge_descriptor hd,
-                    halfedges_around_target(h, mesh_))
+      std::vector<Vector_3> normals_patch1;
+      std::vector<Vector_3> normals_patch2;
+      Patch_id patch1 = -1, patch2 = -1;
+      BOOST_FOREACH(halfedge_descriptor hd, hedges)
       {
         Vector_3 n = compute_normal(face(hd, mesh_));
-        //n can be 0 in the splitting step because we remove degenerate faces
-        //only at the end of the splitting step
-        if(n != CGAL::NULL_VECTOR)
-          normals.push_back(n);
-      }
-      BOOST_FOREACH(halfedge_descriptor hd,
-                    halfedges_around_target(opposite(h, mesh_), mesh_))
-      {
-        Vector_3 n = compute_normal(face(hd, mesh_));
-        if(n != CGAL::NULL_VECTOR)
-          normals.push_back(n);
+        if (n == CGAL::NULL_VECTOR) //for degenerate faces
+          continue;
+        Patch_id pid = get_patch_id(face(hd, mesh_));
+
+        if (patch1 == Patch_id(-1))
+          patch1 = pid; //not met yet
+        else if (patch2 == Patch_id(-1) && patch1 != pid)
+          patch2 = pid; //not met yet
+        CGAL_assertion(pid == patch1 || pid == patch2);
+
+        if (pid == patch1)     normals_patch1.push_back(n);
+        else                   normals_patch2.push_back(n);
       }
 
+      //on each surface patch,
       //check all normals have same orientation
-      bool res = true;
-      for(std::size_t i = 1; i < normals.size(); ++i)/*start at 1 on purpose*/
-      {
-        if (normals[i-1] * normals[i] <= 0.)
-        {
-          res = false;
-          break;
-        }
-      }
+      bool res = check_orientation(normals_patch1)
+              && check_orientation(normals_patch2);
+
       //restore position
       put(vpmap_, vs, ps);
       return res;
@@ -1238,6 +1325,10 @@ private:
     {
       return get(vcmap_, v);
     }
+    void set_constrained(const vertex_descriptor& v, const bool b)
+    {
+      put(vcmap_, v, b);
+    }
 
     bool is_corner(const vertex_descriptor& v) const
     {
@@ -1249,11 +1340,14 @@ private:
         if (nb_incident_features > 2)
           return true;
       }
-      return false;
+      return (nb_incident_features == 1);
     }
 
     Vector_3 compute_normal(const face_descriptor& f) const
     {
+      if (f == boost::graph_traits<PM>::null_face())
+        return CGAL::NULL_VECTOR;
+
       halfedge_descriptor hd = halfedge(f, mesh_);
       typename GeomTraits::Triangle_3
         tr(get(vpmap_, target(hd, mesh_)),
@@ -1275,9 +1369,7 @@ private:
       {
         //being part of the border of the mesh is predominant
         if (is_border(h, mesh_))
-          halfedge_status_map_[h] = MESH_BORDER; //erase previous value if exists
-        else
-          halfedge_status_map_[h] = MESH;
+          set_status(h, MESH_BORDER); //erase previous value if exists
       }
 
       //tag PATCH,       //h and hopp belong to the patch to be remeshed
@@ -1286,7 +1378,7 @@ private:
         BOOST_FOREACH(halfedge_descriptor h,
                       halfedges_around_face(halfedge(f, mesh_), mesh_))
         {
-          halfedge_status_map_[h] = PATCH;
+          set_status(h, PATCH);
         }
       }
 
@@ -1295,22 +1387,21 @@ private:
       //tag PATCH_BORDER,//h belongs to the patch, hopp doesn't
       BOOST_FOREACH(edge_descriptor e, edges(mesh_))
       {
-        if (get(ecmap_, e) || get(border_map, e))
+        if (get(ecmap_, e)
+          || get(border_map, e)
+          || get_patch_id(face(halfedge(e, mesh_), mesh_))
+              != get_patch_id(face(opposite(halfedge(e, mesh_), mesh_), mesh_)))
         {
           //deal with h and hopp for borders that are sharp edges to be preserved
           halfedge_descriptor h = halfedge(e, mesh_);
-          if (halfedge_status_map_[h] == PATCH)
-            halfedge_status_map_[h] = PATCH_BORDER;
+          if (status(h) == PATCH)
+            set_status(h, PATCH_BORDER);
 
           halfedge_descriptor hopp = opposite(h, mesh_);
-          if (halfedge_status_map_[hopp] == PATCH)
-            halfedge_status_map_[hopp] = PATCH_BORDER;
-
-          put(ecmap_, e, false);
+          if (status(hopp) == PATCH)
+            set_status(hopp, PATCH_BORDER);
         }
       }
-
-      CGAL_assertion(halfedge_status_map_.size() == nb_valid_halfedges());
     }
 
     Halfedge_status status(const halfedge_descriptor& h) const
@@ -1319,17 +1410,23 @@ private:
         halfedge_descriptor, Halfedge_status >::const_iterator
           it = halfedge_status_map_.find(h);
       if (it == halfedge_status_map_.end())
-        std::cout << "Something goes wrong with status function" << std::endl;
-      CGAL_assertion(it != halfedge_status_map_.end());
+        return MESH;
       return it->second;
     }
 
+    void set_status(const halfedge_descriptor& h,
+                    const Halfedge_status& s)
+    {
+      if (s == MESH)
+        halfedge_status_map_.erase(h);
+      else
+        halfedge_status_map_[h] = s;
+    }
+
     void merge_status(const halfedge_descriptor& en,
                       const Halfedge_status& s_epo,
                       const Halfedge_status& s_ep)
     {
-      CGAL_assertion(halfedge_status_map_.find(en) != halfedge_status_map_.end());
-
       //get missing data
       halfedge_descriptor eno = opposite(en, mesh_);
       Halfedge_status s_eno = status(eno);
@@ -1342,8 +1439,8 @@ private:
         || s_epo == PATCH_BORDER
         || s_ep == PATCH_BORDER)
       {
-        halfedge_status_map_[en]  = s_epo;
-        halfedge_status_map_[eno] = s_ep;
+        set_status(en, s_epo);
+        set_status(eno, s_ep);
       }
       // else keep current status for en and eno
     }
@@ -1400,13 +1497,13 @@ private:
             CGAL_assertion_code(++nb_done);
 
             //update halfedge_status_map_
-            halfedge_status_map_[h_ab] = merge_status(h_ab, hf, hfo);
-            halfedge_status_map_[h_ca] = merge_status(h_ca, hf, hfo);
-            halfedge_status_map_[hf] =
-              (is_on_patch(h_ca) || is_on_patch_border(h_ca))
-              ? PATCH
-              : MESH;
-            halfedge_status_map_[hfo] = status(hf);
+            set_status(h_ab, merge_status(h_ab, hf, hfo));
+            set_status(h_ca, merge_status(h_ca, hf, hfo));
+            if (is_on_patch(h_ca) || is_on_patch_border(h_ca))
+            {
+              set_status(hf, PATCH);
+              set_status(hfo, PATCH);
+            }
 #ifdef CGAL_PMP_REMESHING_DEBUG
             debug_status_map();
 #endif
@@ -1608,30 +1705,39 @@ private:
     }
 #endif
 
-    //warning : when v is on a sharp edge (angle <= 90 deg)
-    // which is not constrained (it's not mandatory)
-    //this test will return false, though normals are correct
+    //check whether the normals to faces incident to v
+    //have all their 2 by 2 dot products > 0
     bool check_normals(const vertex_descriptor& v) const
     {
-      if (!is_on_patch(v))
-        return true;//not much to say if we are on a boundary/sharp edge
+      if (is_corner(v))
+        return true;//if we want to deal with this case,
+                    //we should use a multimap to store <Patch_id, Normal>
 
-      std::vector<Vector_3> normals;
+      std::vector<Vector_3> normals_patch1;
+      std::vector<Vector_3> normals_patch2;
+      Patch_id patch1 = -1, patch2 = -1;
       BOOST_FOREACH(halfedge_descriptor hd,
                     halfedges_around_target(halfedge(v, mesh_), mesh_))
       {
         Vector_3 n = compute_normal(face(hd, mesh_));
-        normals.push_back(n);
-        CGAL_assertion(n != CGAL::NULL_VECTOR);
+        if (n == CGAL::NULL_VECTOR)
+          continue;
+        Patch_id pid = get_patch_id(face(hd, mesh_));
+
+        if (patch1 == Patch_id(-1))
+          patch1 = pid; //not met yet
+        else if (patch2 == Patch_id(-1) && patch1 != pid)
+          patch2 = pid; //not met yet
+        CGAL_assertion(pid == patch1 || pid == patch2);
+
+        if (pid == patch1)     normals_patch1.push_back(n);
+        else                   normals_patch2.push_back(n);
       }
+
+      //on each surface patch,
       //check all normals have same orientation
-      for (std::size_t i = 1; i < normals.size(); ++i)/*start at 1 on purpose*/
-      {
-        double dot = normals[i - 1] * normals[i];
-        if(dot <= 0.)
-          return false;
-      }
-      return true;
+      return check_orientation(normals_patch1)
+          && check_orientation(normals_patch2);
     }
 
     bool check_normals(const halfedge_descriptor& h) const
@@ -1643,6 +1749,19 @@ private:
       return n * no > 0.;
     }
 
+    bool check_orientation(const std::vector<Vector_3>& normals) const
+    {
+      if (normals.size() < 2)
+        return true;
+      for (std::size_t i = 1; i < normals.size(); ++i)/*start at 1 on purpose*/
+      {
+        double dot = to_double(normals[i - 1] * normals[i]);
+        if (dot <= 0.)
+          return false;
+      }
+      return true;
+    }
+
   public:
     const Triangle_list& input_triangles() const {
       return input_triangles_;
@@ -1673,7 +1792,7 @@ private:
     Patch_id_list input_patch_ids_;
     boost::unordered_map<halfedge_descriptor, Halfedge_status> halfedge_status_map_;
     bool protect_constraints_;
-    boost::unordered_map<face_descriptor, Patch_id> patch_ids_map_;
+    FacePatchMap patch_ids_map_;
     EdgeIsConstrainedMap ecmap_;
     VertexIsConstrainedMap vcmap_;
 
